# Copyright (C) 2017-2018 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import json
import re
import requests

from .cached_objects import CachedObject
from .config.validators import ValidatorASSet
from .errors import PeeringDBError, PeeringDBNoInfoError, ConfigError
from .irrdb import IRRDBInfo


class PeeringDBInfo(CachedObject):

    EXPIRY_TIME_TAG = "pdb_info"

    MISSING_INFO_EXCEPTION = PeeringDBNoInfoError

    def _get_peeringdb_url(self):
        raise NotImplementedError()

    @staticmethod
    def _read_from_url(url):
        response = requests.get(url)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return "{}"
            else:
                raise PeeringDBError(
                    "HTTP error while retrieving info from PeeringDB: "
                    "{}".format(
                        str(e)
                    )
                )
        except Exception as e:
            raise PeeringDBError(
                "Error while retrieving info from PeeringDB: {}".format(
                    str(e)
                )
            )

        return response.content.decode("utf-8")

    def _get_data_from_peeringdb(self):
        plain_text = self._read_from_url(self._get_peeringdb_url())
        try:
            data = json.loads(plain_text)
            return data
        except Exception as e:
            raise PeeringDBError(
                "Error while decoding PeeringDB output: {}".format(
                    str(e)
                )
            )

    def _get_data(self):
        data = self._get_data_from_peeringdb()
        if "data" not in data:
            raise PeeringDBNoInfoError("Missing 'data'")
        if not isinstance(data["data"], list):
            raise PeeringDBNoInfoError("Unexpected format: 'data' is not a list")
        if len(data["data"]) == 0:
            raise PeeringDBNoInfoError("No data for this nextwork")

        return data["data"]

class PeeringDBNet(PeeringDBInfo):

    EXPIRY_TIME_TAG = "pdb_info"

    PEERINGDB_URL = "https://www.peeringdb.com/api/net?asn={asn}"

    def __init__(self, asn, **kwargs):
        PeeringDBInfo.__init__(self, **kwargs)
        self.asn = asn

    def load_data(self):
        logging.debug("Getting data from PeeringDB: net {}".format(self.asn))

        PeeringDBInfo.load_data(self)

        self.info_prefixes4 = self.raw_data[0].get("info_prefixes4", None)
        self.info_prefixes6 = self.raw_data[0].get("info_prefixes6", None)
        self.irr_as_sets = self.parse_as_sets(
            self.raw_data[0].get("irr_as_set", None)
        )

    def parse_as_sets(self, raw_irr_as_sets):
        res = []
        if raw_irr_as_sets and raw_irr_as_sets.strip():
            raw_irr_as_sets = re.split("[/,&\s]", raw_irr_as_sets)
            for raw_irr_as_set in raw_irr_as_sets:
                irr_as_set = self.parse_as_set(raw_irr_as_set)
                if irr_as_set and irr_as_set not in res:
                    res.append(irr_as_set)
        return res

    def parse_as_set(self, in_value):
        v = in_value.strip()

        if not v:
            return None

        guessed = False

        # Removing "ipv4:" and "ipv6:".
        pattern = re.compile("^(?:ipv4|ipv6):", flags=re.IGNORECASE)
        v, number_of_subs_made = pattern.subn("", v)
        if number_of_subs_made > 0:
            v = v.strip()
            guessed = True
        if not v:
            return None

        # IRR record source
        valid_sources_regex = IRRDBInfo.BGPQ3_DEFAULT_SOURCES.replace(",", "|")

        # Converting stuff like AS-FOO@SOURCE in SOURCE::AS-FOO
        pattern = re.compile(
            "^([^@]+)@({sources})$".format(sources=valid_sources_regex),
            flags=re.IGNORECASE
        )
        v, number_of_subs_made = pattern.subn("\\2::\\1", v)
        if number_of_subs_made > 0:
            guessed = True

        # Converting "SOURCE:AS-FOO" format (single colon) to "SOURCE::AS-FOO"
        # (only for known sources)
        pattern = re.compile(
            "^({sources}):([^:].+)$".format(sources=valid_sources_regex),
            flags=re.IGNORECASE
        )
        v, number_of_subs_made = pattern.subn("\\1::\\2", v)
        if number_of_subs_made > 0:
            guessed = True

        try:
            v = ValidatorASSet().validate(v)
        except ConfigError as e:
            logging.debug("AS-SET from PeeringDB for AS{}: "
                          "ignoring {}, {}".format(self.asn, v, str(e)))
            return None

        if guessed:
            logging.info("AS-SET from PeeringDB for AS{}: "
                         "guessed {} from {}".format(self.asn, v, in_value))

        return v

    def _get_object_filename(self):
        return "peeringdb_net_{}.json".format(self.asn)

    def _get_peeringdb_url(self):
        return self.PEERINGDB_URL.format(asn=self.asn)

class PeeringDBNetIXLan(PeeringDBInfo):

    PEERINGDB_URL = "https://www.peeringdb.com/api/netixlan?ixlan_id={ixlanid}"

    def __init__(self, ixlanid, **kwargs):
        PeeringDBInfo.__init__(self, **kwargs)
        self.ixlanid = ixlanid

    def load_data(self):
        logging.debug("Getting data from PeeringDB: Net IX LAN {}".format(self.ixlanid))

        PeeringDBInfo.load_data(self)

    def _get_object_filename(self):
        return "peeringdb_ixlanid_{}.json".format(self.ixlanid)

    def _get_peeringdb_url(self):
        return self.PEERINGDB_URL.format(ixlanid=self.ixlanid)

class PeeringDBIXList(PeeringDBInfo):

    PEERINGDB_URL = "https://peeringdb.com/api/ix"

    def __init__(self, **kwargs):
        PeeringDBInfo.__init__(self, **kwargs)

        self.ixp_list = []

    def load_data(self):
        logging.debug("Getting the list of IXs from PeeringDB")

        PeeringDBInfo.load_data(self)

        for ixp in self.raw_data:
            self.ixp_list.append({
                "city": ixp["city"],
                "country": ixp["country"],
                "full_name": ixp["name_long"],
                "short_name": ixp["name"],
                "peeringdb_handle": ixp["id"]
            })

    def _get_object_filename(self):
        return "peeringdb_ixlist.json"

    def _get_peeringdb_url(self):
        return self.PEERINGDB_URL

def clients_from_peeringdb(netixlanid, cache_dir):
    clients = []

    pdb_net_ixlan = PeeringDBNetIXLan(netixlanid, cache_dir=cache_dir)
    pdb_net_ixlan.load_data()
    netixlans = pdb_net_ixlan.raw_data
    for netixlan in netixlans:
        if netixlan["is_rs_peer"] is True:
            client = {
                "asn": netixlan["asn"],
                "ip": [],
            }
            for ipver in ("ipaddr4", "ipaddr6"):
                if netixlan[ipver]:
                    client["ip"].append(netixlan[ipver].encode("ascii", "ignore").decode("utf-8"))
            clients.append(client)

    asns = {}

    for client in clients:
        asn = client["asn"]
        net = PeeringDBNet(asn, cache_dir=cache_dir)
        net.load_data()

        if not net.irr_as_sets:
            continue

        key = "AS{}".format(asn)
        if key not in asns:
            asns[key] = {
                "as_sets": []
            }

        for irr_as_set in net.irr_as_sets:
            irr_as_set = irr_as_set.strip()
            if irr_as_set not in asns[key]["as_sets"]:
                asns[key]["as_sets"].append(irr_as_set.encode("ascii", "ignore").decode("utf-8"))

    data = {
        "asns": asns,
        "clients": clients
    }

    return data



# Copyright (C) 2017-2018 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import time

from .instances import InstanceError, BGPSpeakerInstance, InstanceNotRunning

class DockerInstance(BGPSpeakerInstance):

    DOCKER_PATH = "docker"
    DOCKER_INSTANCE_PREFIX = "ars_"
    DOCKER_NETWORK_NAME = "arouteserver"
    DOCKER_NETWORK_SUBNET_IPv4 = "192.0.2.0/24"
    DOCKER_NETWORK_SUBNET_IPv6 = "2001:db8:1:1::/64"

    def __init__(self, *args, **kwargs):
        super(DockerInstance, self).__init__(*args, **kwargs)
        self.image = self.DOCKER_IMAGE

    @classmethod
    def _run(cls, cmd, detached=False):
        try:
            if detached:
                dev_null = open(os.devnull, "w")
                subprocess.Popen(
                    cmd.split(),
                    stdin=None,
                    stdout=dev_null,
                    stderr=dev_null
                )
                return None
            else:
                stdout = subprocess.check_output(cmd.split()).decode("utf-8")
                return stdout
        except subprocess.CalledProcessError as e:
            raise InstanceError(
                "Error executing the following command:\n"
                "\t{}\n"
                "Output follows:\n\n"
                "{}".format(cmd, e.output)
            )

    @classmethod
    def _instance_is_running(cls, name):
        cmd = '{docker} ps -f name={prefix}{name} --format="{{{{.ID}}}}"'.format(
            docker=cls.DOCKER_PATH,
            prefix=cls.DOCKER_INSTANCE_PREFIX, name=name
        )
        res = cls._run(cmd)
        if not res:
            return False
        else:
            return True

    def _docker_image_exists(self):
        cmd = '{docker} images {image} --format="{{{{.Repository}}}}'.format(
            docker=self.DOCKER_PATH,
            image=self.image
        )
        res = self._run(cmd)
        if res is None:
            raise InstanceError("Can't get the list of Docker images")
        else:
            return len(res.strip()) > 0

    @classmethod
    def _setup_networking(cls):
        network_details = None
        try:
            network_details = cls._run("{} network inspect {}".format(
                cls.DOCKER_PATH,
                cls.DOCKER_NETWORK_NAME)
            )
        except:
            # network does not exist
            pass

        if network_details is not None:
            try:
                if cls.DOCKER_NETWORK_SUBNET_IPv4 not in network_details:
                    raise InstanceError("IPv4")
                if cls.DOCKER_NETWORK_SUBNET_IPv6 not in network_details:
                    raise InstanceError("IPv6")
            except InstanceError as e:
                raise InstanceError(
                    "The Docker network used by ARouteServer live tests "
                    "('{net}') already exists but is on a "
                    "wrong {v} subnet. Plase consider removing it "
                    "with '{docker} network rm {net}'.".format(
                        net=cls.DOCKER_NETWORK_NAME,
                        v=str(e),
                        docker=cls.DOCKER_PATH
                    )
                )
            return

        try:
            cls._run("{} network create --ipv6 --subnet={} --subnet={} {}".format(
                cls.DOCKER_PATH,
                cls.DOCKER_NETWORK_SUBNET_IPv4,
                cls.DOCKER_NETWORK_SUBNET_IPv6,
                cls.DOCKER_NETWORK_NAME
            ))
        except Exception as e:
            raise InstanceError(
                "Error while creating Docker network '{}': {}.".format(
                    cls.DOCKER_NETWORK_NAME, str(e)
                )
            )

    def is_running(self):
        return self._instance_is_running(self.name)

    def _get_start_cmd(self):
        raise NotImplementedError()

    def start(self):
        self._setup_networking()

        if not self.is_running():

            if not self._docker_image_exists():
                raise InstanceError(
                    "Docker image '{image}' is not present on this system. "
                    "Build it using "
                    "'docker build -t {image} -f PATH_TO_DOCKERFILE .' "
                    "or pull it from DockerHub using "
                    "'docker pull {image}'.".format(image=self.image)
                )

            cmd = ('{docker} run --rm '
                   '--net={net_name} {ip_arg}={ip} '
                   '--name={prefix}{name} '
                   '{mounts} {image} {start_cmd}'.format(
                        docker=self.DOCKER_PATH,
                        net_name=self.DOCKER_NETWORK_NAME,
                        ip_arg="--ip6" if ":" in self.ip else "--ip",
                        ip=self.ip,
                        name=self.name,
                        prefix=self.DOCKER_INSTANCE_PREFIX,
                        mounts=" ".join([
                            "-v{host}:{container}".format(
                                host=mount["var_path"],
                                container=mount["container"]
                            )
                            for mount in self.get_mounts()
                        ]),
                        image=self.image,
                        start_cmd=self._get_start_cmd()
                    )
            )

            self._run(cmd, detached=True)
            time.sleep(3)
            if not self.is_running():
                process = subprocess.Popen(
                    cmd.split(),
                    stdin=None,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE
                )
                stdout, stderr = process.communicate()
                print("STDOUT:")
                print(stdout)
                print("STDERR:")
                print(stderr)

                raise InstanceError(
                    "Can't run detached instance: {}\n"
                    "cmd:\n"
                    "{}".format(self.name, cmd)
                )
        else:
            raise InstanceError("Instance '{}' already running.".format(self.name))

    def stop(self):
        if not self.is_running():
            return

        cmd = '{docker} stop --time=2 {prefix}{name}'.format(
            docker=self.DOCKER_PATH,
            prefix=self.DOCKER_INSTANCE_PREFIX,
            name=self.name
        )
        res = self._run(cmd)
        return res

    def run_cmd(self, args):
        if not self.is_running():
            raise InstanceNotRunning(self.name)

        cmd = '{docker} exec -it {prefix}{name} {args}'.format(
            docker=self.DOCKER_PATH,
            prefix=self.DOCKER_INSTANCE_PREFIX,
            name=self.name,
            args=" ".join(args) if isinstance(args, list) else args
        )
        res = self._run(cmd)
        return res



# Copyright (C) 2017-2018 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import sys
import unittest


class CaptureLog(logging.Handler):

    def __init__(self, *args, **kwargs):
        self.reset_messages()
        super(CaptureLog, self).__init__(*args, **kwargs)

    def reset_messages(self):
        self.msgs = []
        self.warnings = []

    def emit(self, record):
        self.acquire()
        try:
            if record.levelname.lower() == "error":
                self.msgs.append(record.getMessage())
            if record.levelname.lower() == "warning":
                self.warnings.append(record.getMessage())
        finally:
            self.release()

    def reset(self):
        self.acquire()
        try:
            self._reset_messages()
        finally:
            self.release()

class ARouteServerTestCase(unittest.TestCase):

    NEED_TO_CAPTURE_LOG = False
    SHORT_DESCR = ""
    DEBUG = False
    SKIP_ON_TRAVIS = False

    def _capture_log(self):
        self.logger_handler = None
        if self.NEED_TO_CAPTURE_LOG:
            logger = logging.getLogger()
            self.logger_handler = CaptureLog(level="DEBUG")
            logger.addHandler(self.logger_handler)

            self.logger_handler.reset_messages()

    def clear_log(self):
        if self.logger_handler:
            self.logger_handler.reset_messages()

    def _setUp(self):
        pass

    def setUp(self):
        self._capture_log()
        self._setUp()

    @classmethod
    def _setUpClass(cls):
        pass

    @classmethod
    def setUpClass(cls):
        cls._setUpClass()

    @classmethod
    def _tearDownClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        cls._tearDownClass()

    @classmethod
    def print_msg(cls, s):
        sys.stderr.write("{}\n".format(s))

    @classmethod
    def debug(cls, s):
        if cls.DEBUG or "DEBUG" in os.environ:
            cls.print_msg("DEBUG: {}".format(s))

    @classmethod
    def info(cls, s):
        cls.print_msg(s)

    def shortDescription(self):
        return self._testMethodDoc.format(self.SHORT_DESCR)



# Copyright (C) 2017-2018 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from bz2 import decompress
import json
import logging
import os
import requests
from packaging import version

from .ipaddresses import IPNetwork
from .cached_objects import CachedObject
from .errors import ARINWhoisDBDumpError


class ARINWhoisDBDump(CachedObject):

    EXPIRY_TIME_TAG = "arin_whois_db_dump"

    def __init__(self, *args, **kwargs):
        CachedObject.__init__(self, *args, **kwargs)

        self.source = kwargs.get("source")

        self.whois_records = []

    def load_data(self):
        CachedObject.load_data(self)

        logging.debug("Processing ARIN Whois DB dump")

        dic = self.raw_data

        try:
            json_schema = dic.get("json_schema", None)
            if json_schema is None:
                raise ValueError("'json_schema' key is missing")
            if version.parse(json_schema) >= version.parse("0.2"):
                raise ValueError(
                    "unsupported JSON schema version: {}".format(json_schema)
                )

            source = dic.get("source", None)
            if source is None:
                raise ValueError("'source' key is missing")
            if source != "ARIN-WHOIS":
                raise ValueError(
                    "unsupported source: {}".format(source)
                )

            whois_records = dic.get("whois_records", None)
            if whois_records is None:
                raise ValueError("'whois_records' key is missing")
            if not isinstance(whois_records, dict):
                raise ValueError("'whois_records' is not a list")
            if "v4" not in whois_records and "v6" not in whois_records:
                raise ValueError("'v4' and 'v6' lists missing")

            for v4_v6 in ["v4", "v6"]:
                if not isinstance(whois_records.get(v4_v6, []), list):
                    raise ValueError("'{}': a list was expected".format(
                        v4_v6))
                for record in whois_records.get(v4_v6, []):
                    try:
                        if not isinstance(record, dict):
                            raise ValueError("a dict was expected")
                        if "originas" not in record:
                            raise ValueError("'originas' key is missing")
                        originas = record["originas"]
                        if not originas.startswith("AS"):
                            raise ValueError("Origin AS must start with 'AS'")
                        if not originas[2:].isdigit():
                            raise ValueError(
                                "Origin AS must be in 'AS<n>' format"
                            )
                        if "prefix" not in record:
                            raise ValueError("'prefix' key is missing")
                        prefix = record["prefix"]
                        try:
                            IPNetwork(prefix)
                        except Exception as e:
                            raise ValueError("invalid prefix: {} - {}".format(
                                prefix, str(e)
                            ))
                        self.whois_records.append({
                            "originas": record["originas"],
                            "prefix": record["prefix"]
                        })
                    except ValueError as e:
                        raise ValueError(
                            "invalid record '{}': {}".format(
                                str(record), str(e)
                            )
                        )
        except ValueError as e:
            msg = (
                "An error occurred while processing the ARIN Whois "
                "database dump: {}".format(str(e))
            )
            if self.from_cache:
                logging.warning("{} - trying to bypass the cache".format(msg))
                self.bypass_cache = True
                self.load_data()
            else:
                raise ARINWhoisDBDumpError(msg)

    def _get_object_filename(self):
        return "arin-whois-db-dump.json"

    def _get_data(self):
        if self.source.lower().startswith("http://") or \
            self.source.lower().starswith("https://"):

            logging.debug("Downloading ARIN Whois DB dump")

            url = self.source
            try:
                response = requests.get(url).content
            except requests.exceptions.HTTPError as e:
                raise ARINWhoisDBDumpError(
                    "HTTP error while retrieving ARIN Whois DB dump "
                    "from {}: {}".format(
                        url, str(e)
                    )
                )
            except Exception as e:
                raise ARINWhoisDBDumpError(
                    "Error while retrieving ARIN Whois DB dump "
                    "from {}: {}".format(
                        url, str(e)
                    )
                )
        else:
            logging.debug("Loading ARIN Whois DB dump")

            path = self.source
            if not os.path.exists(path):
                raise ARINWhoisDBDumpError(
                    "The ARIN Whois DB dump can't be found, "
                    "file not found: {}".format(path)
                )
            try:
                with open(path, "rb") as f:
                    response = f.read()
            except Exception as e:
                raise ARINWhoisDBDumpError(
                    "Error while reading the ARIN Whois DB dump "
                    "from {}: {}".format(path, str(e))
                )

        if self.source.endswith(".bz2"):
            try:
                raw = decompress(response).decode("utf-8")
            except Exception as e:
                raise ARINWhoisDBDumpError(
                    "An error occurred while "
                    "decompressing ARIN Whois DB "
                    "BZ2 file: {}".format(str(e))
                )
        else:
            raw = response.decode("utf-8")

        try:
            ret = json.loads(raw)
        except Exception as e:
            raise ARINWhoisDBDumpError(
                "Can't parse ARIN Whois DB JSON file: {}".format(str(e))
            )

        return ret



# Copyright (C) 2017-2018 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import yaml

from .base import ARouteServerCommand

from ..config.base import convert_deprecated
from ..config.general import ConfigParserGeneral
from ..config.program import program_config

class ShowConfigCommand(ARouteServerCommand):

    NEEDS_CONFIG = True

    COMMAND_NAME = "show_config"
    COMMAND_HELP = ("Show the current general configuration "
                    "settings and highlight options which are "
                    "missing or left to their default settings.")

    @classmethod
    def add_arguments(cls, parser):
        super(ShowConfigCommand, cls).add_arguments(parser)

        parser.add_argument(
            "--general",
            help="General route server configuration file.",
            metavar="FILE",
            dest="cfg_general")

    def run(self):
        current_config_path = program_config.get("cfg_general")
        self.show_config(current_config_path, sys.stdout)

    @staticmethod
    def show_config(current_config_path, output):

        def wr_line(configured, level, line):
            if configured is True:
                status = "configured"
            elif configured is False:
                status = "default"
            else:
                status = ""

            output.write("{status:<15} {indent}{line}\n".format(
                status=status,
                indent="  " * level,
                line=line
            ))

        with open(current_config_path, "r") as f:
            current_config = yaml.safe_load(f)
        convert_deprecated(current_config["cfg"])

        distrib = ConfigParserGeneral()
        distrib._load_from_yaml("cfg:\n"
                                "  rs_as: 65534\n"
                                "  router_id: 192.0.2.1\n")
        distrib.parse()
        del distrib["communities"]
        del distrib["custom_communities"]

        ordered_schema = ConfigParserGeneral.get_schema()
        del ordered_schema["cfg"]["communities"]
        del ordered_schema["cfg"]["custom_communities"]

        def get_val_repr(val):
            if isinstance(val, dict):
                return ", ".join("{}: {}".format(k, v) for k, v in sorted(val.items()))
            else:
                return str(val).strip()

        def iterate(ordered_schema, distrib, current, level=0):
            for k in ordered_schema:
                if current is distrib:
                    configured = False
                else:
                    configured = k in current

                iterate_over = current if configured else distrib

                if isinstance(ordered_schema[k], dict):

                    wr_line(None, level, k + ":")
                    iterate(ordered_schema[k], distrib[k], iterate_over[k],
                            level + 1)

                elif isinstance(ordered_schema[k], list) or \
                    isinstance(iterate_over[k], list):

                    wr_line(configured, level, k + ":")
                    for v in iterate_over[k]:
                        wr_line(configured, level, "  - " + get_val_repr(v))

                else:
                    v = get_val_repr(iterate_over[k])
                    wr_line(configured, level, "{}: {}".format(k, v))

        iterate(ordered_schema, distrib.cfg, current_config)



# Copyright (C) 2017-2018 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os

from .base import BaseConfigEnricher
from ..ipaddresses import IPNetwork
from ..errors import ARouteServerError, BuilderError

class GenericIRRWhoisRecord_Proxy(object):

    def __init__(self, asn, path, allow_longer_prefixes):
        self.asn = asn
        self.path = path
        self.allow_longer_prefixes = allow_longer_prefixes

    @property
    def prefixes(self):
        with open(self.path, "r") as f:
            prefix_lst = json.load(f)
            for prefix in prefix_lst:
                net = IPNetwork(prefix)
                yield {
                    "prefix": net.ip,
                    "length": net.prefixlen,
                    "max_length": net.max_prefixlen,
                    "exact": not self.allow_longer_prefixes,
                    "ge": net.prefixlen,
                    "le": net.prefixlen if not self.allow_longer_prefixes else net.max_prefixlen
                }

class GenericIRRWhoisDBDumpEnricher(BaseConfigEnricher):

    DESCR = None
    DIR_NAME = None
    CONFIG_SECTION_NAME = None
    PARSER_CLASS = None
    BUILDER_TARGET_DICT_NAME = None

    def enrich(self):
        if self.builder.irrdb_info is None:
            raise BuilderError(
                "{} Whois DB records can be fetched only after that the "
                "list of authorized origin ASNs has been built.".format(
                    self.DESCR
                )
            )

        # List of all the origin ASNs.
        origin_asns = set()
        for bundle_id in self.builder.irrdb_info:
            bundle = self.builder.irrdb_info[bundle_id]
            origin_asns.update(bundle.asns)

        if not origin_asns:
            return

        logging.info("Updating entries from the {} Whois DB dump...".format(
            self.DESCR))

        cache_dir = self.builder.cache_dir

        db_dir = os.path.join(cache_dir, self.DIR_NAME)
        if not os.path.exists(db_dir):
            try:
                os.makedirs(db_dir)
            except OSError as e:
                raise ARouteServerError(str(e))

        afis = [4, 6] if self.builder.ip_ver is None else [self.builder.ip_ver]

        irrdb_cfg = self.builder.cfg_general["filtering"]["irrdb"]
        whois_db_dump = irrdb_cfg[self.CONFIG_SECTION_NAME]
        source = whois_db_dump["source"]

        whois_db_dump = self.PARSER_CLASS(
            cache_dir=cache_dir, cache_expiry=self.builder.cache_expiry,
            source=source)
        whois_db_dump.load_data()
        whois_records = whois_db_dump.whois_records

        # "ASx": ["1.2.3.0/24", ...]
        asn_prefixes = {}

        for record in whois_records:
            asn = record["originas"]
            prefix = record["prefix"]

            # If the current prefix is for an origin ASN that is
            # not allowed for any client then skip it.
            if int(asn[2:]) not in origin_asns:
                continue

            prefix_obj = IPNetwork(prefix)
            if prefix_obj.version not in afis:
                continue

            asn = asn.upper()
            if asn not in asn_prefixes:
                asn_prefixes[asn] = set()
            asn_prefixes[asn].add(prefix)

        allow_longer_prefixes = self.builder.cfg_general["filtering"]["irrdb"]["allow_longer_prefixes"]
        for asn in asn_prefixes:
            path = os.path.join(db_dir, "{}.json".format(asn))
            with open(path, "w") as f:
                json.dump(list(asn_prefixes[asn]), f)
            target_dic = getattr(self.builder, self.BUILDER_TARGET_DICT_NAME)
            target_dic[asn] = \
                GenericIRRWhoisRecord_Proxy(asn, path, allow_longer_prefixes)

        del asn_prefixes
        del whois_records



# Copyright (C) 2017-2018 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import time

from .instances import InstanceError, BGPSpeakerInstance, InstanceNotRunning

class KVMInstance(BGPSpeakerInstance):

    MAX_BOOT_TIME = 60
    VIRSH_DOMAINNAME = None

    SSH_USERNAME = "root"
    SSH_KEY_PATH = "~/.ssh/arouteserver"

    @classmethod
    def _get_ssh_user(cls):
        if "SSH_USERNAME" in os.environ:
            return os.environ["SSH_USERNAME"]
        return cls.SSH_USERNAME

    @classmethod
    def _get_ssh_key_path(cls):
        if "SSH_KEY_PATH" in os.environ:
            path = os.environ["SSH_KEY_PATH"]
        else:
            path = cls.SSH_KEY_PATH
        return os.path.expanduser(path)

    @classmethod
    def _get_virsh_domainname(cls):
        if "VIRSH_DOMAINNAME" in os.environ:
            return os.environ["VIRSH_DOMAINNAME"]
        return cls.VIRSH_DOMAINNAME

    def __init__(self, *args, **kwargs):
        super(KVMInstance, self).__init__(*args, **kwargs)
        self.domain_name = self._get_virsh_domainname()

        # If set, the VM is always considered up and SSH connections
        # are established toward this IP address.
        self.remote_ip = kwargs.get("remote_ip", None)
        if self.remote_ip:
            self.remote_ip = self.remote_ip.strip()
        self.is_remote = self.remote_ip != "" and self.remote_ip

    @classmethod
    def _run(cls, cmd):
        cls.debug("Executing '{}'".format(cmd))
        try:
            dev_null = open(os.devnull, "w")
            stdout = subprocess.check_output(
                cmd.split(), stderr=dev_null
            ).decode("utf-8")
            return stdout
        except subprocess.CalledProcessError as e:
            raise InstanceError(
                "Error executing the following command:\n"
                "\t{}\n"
                "Output follows:\n\n"
                "{}".format(cmd, e.output)
            )

    def is_running(self):
        if self.is_remote:
            return True

        res = self._run("virsh list --name --state-running")
        return self.domain_name in res

    def _check_env(self):
        if not self.is_remote:
            vms_list_raw = self._run("virsh list --name --all")
            vms_list = vms_list_raw.split("\n")
            found = False
            for vm in vms_list:
                if vm.strip() == self.domain_name:
                    found = True
                    break
            if not found:
                raise Exception(
                    "The virsh domain '{}' does not appear to "
                    "be in the list of configured domains: "
                    "'virsh list --all'. Please check that the KVM "
                    "virtual machine used by the live test framework "
                    "is configured correctly. To use a different "
                    "VM name, set the VIRSH_DOMAINNAME environment "
                    "variable before running the tests.".format(
                        self.domain_name
                    )
                )

        key_file = self._get_ssh_key_path()
        if not os.path.exists(key_file) or not os.path.isfile(key_file):
            raise Exception(
                "The SSH key file needed to connect to the "
                "virtual machine used by the live test framework "
                "does not exist: {}. To use a different path, "
                "set the SSH_USERNAME environment variable before "
                "running the tests.".format(key_file)
            )

    def start(self):
        need_to_start = True

        if self.is_remote:
            need_to_start = False
        else:
            if self.is_running():
                if "REUSE_KVM_INSTANCES" in os.environ:
                    need_to_start = False
                else:
                    raise InstanceError("Instance '{}' already running.".format(self.name))

        if need_to_start:
            self._check_env()

            res = self._run("virsh start {}".format(self.domain_name))

            if "error:" in res:
                raise InstanceError(
                    "Can't run instance '{}'; "
                    "an error occurred while starting virsh domain '{}':\n"
                    "{}".format(self.name, self.domain_name, res)
                )
            time.sleep(3)

            if not self.is_running():
                raise InstanceError(
                    "The virsh domain '{}' failed to start".format(
                        self.domain_name
                    )
                )

            running = False
            for i in range(self.MAX_BOOT_TIME // 2):
                time.sleep(2)
                try:
                    res = self.run_cmd("true")
                    running = True
                    break
                except:
                    pass

            if not running:
                raise InstanceError(
                    "Instance '{}' seems not running after {} seconds".format(
                        self.name, self.MAX_BOOT_TIME
                    )
                )

        # The VM is now up and bgpd should be started too,
        # but configuration files have not been updated yet,
        # so call restart() in order to push them
        # to the VM and restart the daemon.
        self.restart()

    def _graceful_shutdown(self):
        return False

    def stop(self):
        if self.is_remote:
            return

        if not self.is_running():
            return

        if "REUSE_KVM_INSTANCES" in os.environ:
            return

        self._graceful_shutdown()

        for i in range(30 // 2):
            time.sleep(2)
            if not self.is_running():
                return

        if self.is_running() and not self.is_remote:
            try:
                self._run("virsh shutdown {}".format(self.domain_name))
                for i in range(30 // 2):
                    time.sleep(2)
                    if not self.is_running():
                        return
            except:
                pass

        if self.is_running() and not self.is_remote:
            try:
                self._run("virsh destroy {}".format(self.domain_name))
            except:
                pass

    def run_cmd(self, args):
        if not self.is_running():
            raise InstanceNotRunning(self.name)

        cmd = ("ssh -o BatchMode=yes -o ConnectTimeout=5 "
               "-o StrictHostKeyChecking=no "
               "-o ServerAliveInterval=10 {user}@{ip} -i {path_to_key} "
               "{cmd}").format(
            user=self._get_ssh_user(),
            ip=self.remote_ip if self.is_remote else self.ip,
            path_to_key=self._get_ssh_key_path(),
            cmd=" ".join(args) if isinstance(args, list) else args
        )
        res = self._run(cmd)
        return res

    def _mount_files(self):
        for mount in self.get_mounts():
            ip = self.remote_ip if self.is_remote else self.ip
            local_file = mount["host"]
            remote_file = mount["container"]

            gzipped = local_file.endswith(".gz")
            if gzipped:
                remote_file += ".gz"

            cmd = ("scp -i {path_to_key} "
                   "-o StrictHostKeyChecking=no "
                   "{host_file} {user}@{ip}:{container_file} ".format(
                       host_file=local_file,
                       user=self._get_ssh_user(),
                       ip="[{}]".format(ip) if ":" in ip else ip,
                       container_file=remote_file,
                       path_to_key=self._get_ssh_key_path()
                    ))
            self._run(cmd)

            if gzipped:
                self.run_cmd("gunzip -f {}".format(remote_file))



# Copyright (C) 2017-2018 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import json
import re
import requests
import six

from .peering_db import PeeringDBNet, PeeringDBNoInfoError
from .errors import EuroIXError, EuroIXSchemaError

class EuroIXMemberList(object):

    TESTED_EUROIX_SCHEMA_VERSIONS = ("0.4", "0.5", "0.6")

    CUSTOM_COMMUNITIES = ["switch_id", "switch_name", "colocation", "city",
                          "country", "member_type"]
    MEMBER_TYPES = ["peering", "ixp", "routeserver", "probono", "other"]
    EUROIX_SWITCH_ATTRIBUTES_COMMUNITIES_MAP = [
        ("id", "switch_id"),
        ("name", "switch_name"),
        ("colo", "colocation"),
        ("city", "city"),
        ("country", "country")
    ]
    INFO_FROM_PEERINGDB = ["as-set", "max-prefix"]

    def __init__(self, input_object, cache_dir, cache_expiry):
        self.cache_dir = cache_dir
        self.cache_expiry = cache_expiry

        self.raw_data = None

        if isinstance(input_object, dict):
            self.raw_data = input_object
        elif isinstance(input_object, six.string_types):
            response = requests.get(input_object)
            raw = response.content.decode("utf-8")
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as e:
                raise EuroIXError(
                    "Error while retrieving Euro-IX "
                    "JSON file from {}: {}".format(
                        input_object, str(e)
                    )
                )
        else:
            raw = input_object.read().decode("utf-8")

        if not self.raw_data:
            try:
                self.raw_data = json.loads(raw)
            except Exception as e:
                raise EuroIXSchemaError(
                    "Error while processing JSON data: {}".format(str(e))
                )

        self.switches = None
        self.unique_custom_communities = None

        self.check_schema_version()

    @staticmethod
    def _check_type(v, vname, expected_type):
        if expected_type is str:
            expected_type_set = six.string_types
        else:
            expected_type_set = expected_type

        if not isinstance(v, expected_type_set):
            if expected_type is int and \
                isinstance(v, six.string_types) and \
                v.isdigit():
                return int(v)

            raise EuroIXSchemaError(
                "Invalid type for {} with value '{}': "
                "it is {}, should be {}".format(
                    vname, v, str(type(v)), str(expected_type)
                )
            )

        return v

    @staticmethod
    def _get_item(key, src, expected_type=None, optional=False):
        if key not in src:
            if optional:
                return None
            raise EuroIXSchemaError("Missing required item: {}".format(key))
        val = src[key]
        if expected_type:
            val = EuroIXMemberList._check_type(val, key, expected_type)
        return val

    @staticmethod
    def mk_parents_and_set(d, p, v=None):
        if v is not None:
            parents = p.split(".")[:-1]
            val_key = p.split(".")[-1]
        else:
            parents = p.split(".")
            val_key = None

        last = d
        for key in parents:
            if key not in last:
                last[key] = {}
            last = last[key]

        if val_key:
            last[val_key] = v

    def check_schema_version(self):
        data = self.raw_data

        version = self._get_item("version", data, str)
        tested_versions = self.TESTED_EUROIX_SCHEMA_VERSIONS
        if version not in tested_versions:
            logging.warning("The version of the JSON schema of this file ({}) "
                            "is not one of those tested ({}). Unexpected "
                            "errors may occurr.".format(
                                version,
                                ", ".join(tested_versions)
                            ))

    def get_clients(self, ixp_id, vlan_id=None,
                    routeserver_only=False,
                    guess_custom_bgp_communities=[],
                    merge_from_peeringdb=[]):

        def new_client(asn, description):
            client = {}
            client["asn"] = asn
            if description:
                client["description"] = description.encode("ascii", "replace")
                client["description"] = client["description"].strip().decode("utf-8")
            return client

        def get_descr(member, connection=None):

            res = []

            for info, k in [("AS{}", "asnum"),
                            ("name '{}'", "name")]:
                try:
                    res.append(info.format(member[k]))
                except:
                    pass

            if connection:
                for info, k in [("ixp_id {}", "ixp_id"),
                                ("state '{}'", "state")]:
                    try:
                        res.append(info.format(connection[k]))
                    except:
                        pass

                try:
                    for iface in connection["if_list"]:
                        res.append(
                            "iface on switch_id {}".format(iface["switch_id"])
                        )
                except:
                    pass

            if res:
                return ", ".join(res)
            else:
                return "unknown"

        def normalize_bgp_community(s):
            res = s.decode("utf-8")
            res = res.lower()
            res = re.sub("[\\/\[ \]+-]", "_", res)
            res = re.sub("[^0-9a-zA-Z_]", "", res)
            return res

        def attach_custom_bgp_community(client, prefix, name):
            community_tag = normalize_bgp_community(name)

            self.mk_parents_and_set(client, "cfg.attach_custom_communities", [])

            if community_tag not in client["cfg"]["attach_custom_communities"]:
                client["cfg"]["attach_custom_communities"].append(
                    "{}_{}".format(prefix, community_tag)
                )

        def enrich_with_custom_bgp_communities(clients, connection):
            if not guess_custom_bgp_communities:
                return

            if_list = self._get_item("if_list", connection, list, True)
            if not if_list:
                return

            for if_list_entry in if_list:
                switch_id = self._get_item("switch_id", if_list_entry, int, True)
                if not switch_id:
                    return

                if not str(switch_id) in self.switches:
                    return

                switch_info = self.switches[str(switch_id)]

                for attribute, prefix in self.EUROIX_SWITCH_ATTRIBUTES_COMMUNITIES_MAP:
                    if prefix not in guess_custom_bgp_communities:
                        continue
                    if attribute not in switch_info:
                        continue
                    attribute_val = switch_info[attribute]

                    for client in clients:
                        attach_custom_bgp_community(client, prefix, attribute_val)

        def enrich_with_peeringdb_info(clients):
            if not merge_from_peeringdb:
                return

            for client in clients:
                fetch_from_pdb = []
                if "as-set" in merge_from_peeringdb:
                    try:
                        if not client["cfg"]["filtering"]["irrdb"]["as_sets"]:
                            raise KeyError()
                    except KeyError:
                        fetch_from_pdb.append("as-set")

                if "max-prefix" in merge_from_peeringdb:
                    for ip_ver in [4, 6]:
                        try:
                            if not client["cfg"]["filtering"]["max_prefix"]["limit_ipv{}".format(ip_ver)]:
                                raise KeyError()
                        except KeyError:
                            fetch_from_pdb.append("max-prefix{}".format(ip_ver))

                if fetch_from_pdb:
                    try:
                        pdb_net = PeeringDBNet(client["asn"],
                                               cache_dir=self.cache_dir,
                                               cache_expiry=self.cache_expiry)
                        pdb_net.load_data()
                    except PeeringDBNoInfoError:
                        continue

                    as_sets = pdb_net.irr_as_sets
                    if as_sets and "as-set" in fetch_from_pdb:
                        self.mk_parents_and_set(
                            client,
                            "cfg.filtering.irrdb.as_sets",
                            as_sets
                        )

                    for ip_ver in [4, 6]:
                        if ip_ver == 4:
                            max_prefix = pdb_net.info_prefixes4
                        else:
                            max_prefix = pdb_net.info_prefixes6

                        if max_prefix and "max-prefix{}".format(ip_ver) in fetch_from_pdb:
                            self.mk_parents_and_set(
                                client,
                                "cfg.filtering.max_prefix.limit_ipv{}".format(ip_ver),
                                max_prefix
                            )

        def process_member(member):
            if self._get_item("member_type", member, str, True) == "routeserver":
                # Member is a route server itself.
                return

            clients = []
            connection_list = self._get_item("connection_list", member, list)

            for connection in connection_list:
                try:
                    new_clients = process_connection(member, connection)
                    if new_clients:
                        enrich_with_custom_bgp_communities(new_clients,
                                                           connection)

                        enrich_with_peeringdb_info(new_clients)

                        for new_client in new_clients:
                            duplicate_found = False
                            for existing_client in clients:
                                if new_client == existing_client:
                                    duplicate_found = True
                                    break
                            if not duplicate_found:
                                clients.append(new_client)
                except EuroIXError as e:
                    if str(e):
                        logging.error(
                            "Error while processing {}: {}".format(
                                get_descr(member, connection), str(e)
                            )
                        )
                    raise EuroIXError()

            return clients

        def process_connection(member, connection):
            self._check_type(connection, "connection", dict)

            if self._get_item("ixp_id", connection, int) != ixp_id:
                return

            # Member has a connection to the selected IXP infrastructure.

            clients = []
            asnum = self._get_item("asnum", member, int)
            name = self._get_item("name", member, str, True)
            member_type = self._get_item("type", member, str, True)

            vlan_list = self._get_item("vlan_list", connection, list, True)

            if vlan_id and not vlan_list:
                # A specific VLAN has been requested but member does not
                # have information about VLANs at all.
                return

            # Workaround for IXP-Manager issue:
            # https://github.com/inex/IXP-Manager/commit/50c3781711ed38e773f86a8f3017d669d18e464d
            if vlan_list and isinstance(vlan_list[0], list):
                vlan_list = vlan_list[0]

            for vlan in vlan_list or []:
                self._check_type(vlan, "vlan", dict)

                if vlan_id and \
                    self._get_item("vlan_id", vlan, int, True) != vlan_id:
                    # This VLAN is not the requested one.
                    continue

                for ip_ver in (4, 6):
                    ipv4_6 = "ipv{}".format(ip_ver)
                    ip_info = self._get_item(ipv4_6, vlan, dict, True)

                    if not ip_info:
                        continue

                    address = self._get_item("address", ip_info, str, True)

                    if not address:
                        continue

                    if routeserver_only:
                        # Members with routeserver attribute == False
                        # are excluded.
                        if self._get_item("routeserver", ip_info, bool, True) is False:
                            continue

                    client = new_client(asnum, name)
                    client["ip"] = address

                    as_macro = self._get_item("as_macro", ip_info, str, True)
                    max_prefix = self._get_item("max_prefix", ip_info, int, True)

                    if as_macro:
                        self.mk_parents_and_set(
                            client, "cfg.filtering.irrdb.as_sets", [as_macro]
                        )
                    if max_prefix:
                        self.mk_parents_and_set(
                            client,
                            "cfg.filtering.max_prefix.limit_ipv{}".format(ip_ver),
                            max_prefix
                        )
                    if guess_custom_bgp_communities and \
                        "member_type" in guess_custom_bgp_communities:
                        if member_type:
                            if member_type not in self.MEMBER_TYPES:
                                raise EuroIXSchemaError(
                                    "Unexpected member type: '{}'".format(
                                        member_type
                                    )
                                )

                            attach_custom_bgp_community(client, "member_type", member_type)

                    clients.append(client)

            return clients

        def get_custom_bgp_comms_data(ixp):
            raw_switches = self._get_item("switch", ixp, list, True)
            if not raw_switches:
                return

            # switches = {
            #   "<switch_id>": {
            #     "<switch_attr_name>": <value>
            #   }
            # }
            self.switches = {}

            # unique_custom_communities = {
            #   "<custom_community_prefix>": set of unique values
            # }
            self.unique_custom_communities = {
                cust_comm_prefix: set()
                for cust_comm_prefix in self.CUSTOM_COMMUNITIES
            }

            if "member_type" in guess_custom_bgp_communities:
                self.unique_custom_communities["member_types"] = set(
                    ["member_type_{}".format(t) for t in self.MEMBER_TYPES]
                )

            for switch in raw_switches:
                switch_id = self._get_item("id", switch, int, True)

                if not switch_id:
                    continue

                self.switches[str(switch_id)] = {}

                for attribute_name, prefix in self.EUROIX_SWITCH_ATTRIBUTES_COMMUNITIES_MAP:
                    attribute_val = self._get_item(attribute_name, switch,
                                                   optional=True)
                    if not attribute_val:
                        continue

                    if isinstance(attribute_val, int):
                        attribute_val = str(attribute_val)
                    attribute_val = attribute_val.encode("ascii", "replace")

                    if prefix in guess_custom_bgp_communities:
                        self.unique_custom_communities[prefix].add(
                            "{}_{}".format(prefix,
                                           normalize_bgp_community(attribute_val)
                                        )
                        )
                    self.switches[str(switch_id)][attribute_name] = attribute_val

        data = self.raw_data

        ixp_list = self._get_item("ixp_list", data, list)
        member_list = self._get_item("member_list", data, list)

        ixp_found = False
        for ixp in ixp_list:
            self._check_type(ixp, "ixp", dict)

            if self._get_item("ixp_id", ixp, int) == ixp_id:
                ixp_found = True
                break

        if not ixp_found:
            raise EuroIXError(
                "IXP ID {} not found".format(ixp_id))

        if guess_custom_bgp_communities:
            get_custom_bgp_comms_data(ixp)

        raw_clients = []
        for member in member_list:
            try:
                self._check_type(member, "member", dict)
                new_clients = process_member(member)
                if new_clients:
                    raw_clients.extend(new_clients)
            except EuroIXError as e:
                if str(e):
                    logging.error(
                        "Error while processing member {}: {}".format(
                            get_descr(member), str(e)
                        )
                    )
                raise EuroIXError()

        return raw_clients

        # TODO: Merge clients with same IRRDB info

    def print_infrastructure_list(self, out_file):
        data = self.raw_data

        ixp_list = self._get_item("ixp_list", data, list)
        for ixp in ixp_list:
            self._check_type(ixp, "ixp", dict)

            ixp_id = self._get_item("ixp_id", ixp, int)
            shortname = self._get_item("shortname", ixp, str)
            name = self._get_item("name", ixp, str, True)
            country = self._get_item("country", ixp, str, True)

            s = "IXP ID {}, short name '{}'".format(ixp_id, shortname)
            if name:
                s += ", '{}'".format(name)
            if country:
                s += ", {}".format(country)
            out_file.write(s + "\n")

            vlans = self._get_item("vlan", ixp, list, True)

            for vlan in vlans or []:
                s = ""

                vlan_id = self._get_item("id", vlan, int, True)
                if vlan_id:
                    s += ", " if s else ""
                    s += "ID {}".format(vlan_id)

                vlan_name = self._get_item("name", vlan, str, True)
                if vlan_name:
                    s += ", " if s else ""
                    s += "name '{}'".format(vlan_name)

                for ip_ver in (4, 6):
                    ipv4_6 = "ipv{}".format(ip_ver)
                    ip_info = self._get_item(ipv4_6, vlan, dict, True)
                    if not ip_info:
                        continue

                    prefix = self._get_item("prefix", ip_info, str, True)
                    mask_length = self._get_item("mask_length", ip_info, int, True)
                    if prefix:
                        s += ", " if s else ""
                        s += "IPv{} prefix {}".format(ip_ver, prefix)
                        if mask_length:
                            s += "/" + str(mask_length)

                if not s:
                    s = "VLAN with no details"
                else:
                    s = "VLAN " + s

                out_file.write(" - " + s + "\n")



# Copyright (C) 2017-2018 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from .base import BaseConfigEnricher, BaseConfigEnricherThread
from ..errors import BuilderError, PeeringDBError, PeeringDBNoInfoError
from ..peering_db import PeeringDBNet

class PeeringDBConfigEnricher_MaxPrefix_WorkerThread(BaseConfigEnricherThread):

    DESCR = "PeeringDB max-prefix"

    def __init__(self, *args, **kwargs):
        BaseConfigEnricherThread.__init__(self, *args, **kwargs)

        self.ip_ver = None
        self.cfg_general = None
        self.cache_dir = None
        self.cache_expiry = None
        self.general_limits = None

    def do_task(self, task):
        asn, _ = task

        try:
            net = PeeringDBNet(asn,
                               cache_dir=self.cache_dir,
                               cache_expiry=self.cache_expiry)
            net.load_data()

            return net.info_prefixes4 or self.general_limits["ipv4"], \
                    net.info_prefixes6 or self.general_limits["ipv6"]

        except PeeringDBNoInfoError:
            # No data found on PeeringDB.
            logging.debug("No data found on PeeringDB "
                          "for AS{} while looking for "
                          "max-prefix limit.".format(asn))
            pass
        except PeeringDBError as e:
            logging.error(
                "An error occurred while retrieving info from PeeringDB "
                "for ASN {}: {}".format(
                    asn, str(e) or "error unknown"
                )
            )
            raise BuilderError()

    @staticmethod
    def _set_client_max_prefix_from_pdb(client, afi, pdb_value):
        max_prefix = client["cfg"]["filtering"]["max_prefix"]
        pdb_incr = max_prefix["peering_db"]["increment"]
        limit = int(round(
                    (pdb_value + pdb_incr["absolute"]) * \
                    (1 + float(pdb_incr["relative"]) / 100)
                ))
        max_prefix["limit_ipv{}".format(afi)] = limit

    def save_data(self, task, data):
        _, clients = task
        limit4, limit6 = data
        for client in clients:
            client_max_prefix = client["cfg"]["filtering"]["max_prefix"]
            if limit4 and not client_max_prefix["limit_ipv4"]:
                self._set_client_max_prefix_from_pdb(client, 4, limit4)
            if limit6 and not client_max_prefix["limit_ipv6"]:
                self._set_client_max_prefix_from_pdb(client, 6, limit6)

class PeeringDBConfigEnricher_MaxPrefix(BaseConfigEnricher):

    WORKER_THREAD_CLASS = PeeringDBConfigEnricher_MaxPrefix_WorkerThread

    def _get_general_limit(self, ip_ver):
        return self.builder.cfg_general["filtering"]["max_prefix"]["general_limit_ipv{}".format(ip_ver)]

    def _config_thread(self, thread):
        thread.ip_ver = self.builder.ip_ver
        thread.cfg_general = self.builder.cfg_general
        thread.cache_dir = self.builder.cache_dir
        thread.cache_expiry = self.builder.cache_expiry
        thread.general_limits = {
            "ipv4": self._get_general_limit(4),
            "ipv6": self._get_general_limit(6)
        }

    def add_tasks(self):
        # "<asn>": <clients>
        tasks = {}

        # Enqueuing tasks.
        for client in self.builder.cfg_clients.cfg["clients"]:
            client_max_prefix = client["cfg"]["filtering"]["max_prefix"]

            if not client_max_prefix["action"]:
                # No max-prefix action given for this client:
                # no needs to know its max-pref limit.
                continue

            afis = [4, 6] if self.builder.ip_ver is None else [self.builder.ip_ver]

            pdb_info_needed = False

            for ip_ver in afis:
                if client_max_prefix["limit_ipv{}".format(ip_ver)]:
                    # Client uses a specific limit:
                    # no needs to gather info from PeeringDB for
                    # the current address family.
                    continue

                if not client_max_prefix["peering_db"]["enabled"]:
                    # PeeringDB disabled for this client:
                    # using general limit.
                    client_max_prefix["limit_ipv{}".format(ip_ver)] = \
                        self._get_general_limit(ip_ver)
                    continue

                pdb_info_needed = True

            if pdb_info_needed:
                asn = str(client["asn"])
                if asn not in tasks:
                    tasks[asn] = []
                tasks[asn].append(client)

        for asn in tasks:
            self.tasks_q.put((int(asn), tasks[asn]))



# Copyright (C) 2017-2018 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import re
import yaml


from .validators import ConfigParserValidator
from ..errors import ConfigError, MissingFileError, ARouteServerError


class ConfigParserBase(object):

    ROOT = None

    def __init__(self):
        self.cfg = None
        self.file_dir = None

    def __contains__(self, name):
        return name in self.cfg[self.ROOT]

    def __getitem__(self, name):
        return self.cfg[self.ROOT][name]

    def __setitem__(self, name, val):
        self.cfg[self.ROOT][name] = val

    def __delitem__(self, name):
        del self.cfg[self.ROOT][name]

    def _load_from_yaml(self, doc):

        def expand_include(lines):
            res = ""
            for line in lines:
                if not line:
                    continue

                if not line.strip():
                    continue

                if res:
                    res += "\n"

                if line.strip().startswith("!include"):
                    filepath = line.strip().split(" ")[1]
                    filepath = os.path.expanduser(filepath)

                    if not os.path.isabs(filepath):
                        if self.file_dir:
                            filepath = os.path.join(self.file_dir, filepath)

                    with open(filepath) as inputfile:
                        raw = inputfile.read()

                    res += expand_include(raw.split("\n"))
                    continue

                res += line

            return res

        def expand_env_vars(doc):
            res = doc
            for v in os.environ:
                res = re.sub("\$\{" + v + "\}", os.environ[v], res)
            res = re.sub("\$\{[A-Za-z0-9_]+\}", "", res)
            return res

        lines = doc.split("\n")
        expanded_doc = expand_include(lines)
        expanded_doc = expand_env_vars(expanded_doc)

        try:
            self.cfg = yaml.safe_load(expanded_doc)
        except Exception as e:
            raise ConfigError(
                "Can't parse YAML file: {}".format(str(e))
            )

        if not isinstance(self.cfg, dict):
            raise ConfigError(
                "Error while parsing config file: invalid syntax. "
                "Hint: check that the root element '{}' exists.".format(
                    self.ROOT
                )
            )

    def _load_from_yaml_file(self, cfg_path):
        if not os.path.isfile(cfg_path):
            raise MissingFileError(cfg_path)

        self.file_dir = os.path.dirname(cfg_path)

        with open(cfg_path, "r") as f:
            self._load_from_yaml(f.read())

    def load(self, cfg_path):
        self._load_from_yaml_file(cfg_path)

        try:
            self.parse()
        except ARouteServerError as e:
            if str(e):
                logging.error(str(e))
            raise ConfigError()

    @staticmethod
    def validate(schema, cfg, path=""):
        errors = False
        if cfg is None:
            cfg = {}
            return

        for prop in cfg:
            if prop not in schema:
                errors = True
                logging.error(
                    "Unknown statement at '{}' level: '{}'.".format(
                        path, prop
                    )
                )

        for prop in schema:
            if isinstance(schema[prop], ConfigParserValidator):
                validator = schema[prop]

                try:
                    if prop in cfg:
                        cfg[prop] = validator.validate(cfg[prop])
                    else:
                        cfg[prop] = validator.validate(None)
                except ConfigError as e:
                    errors = True
                    logging.error(
                        "Error parsing '{}' at '{}' level - {}.".format(
                            prop, path, str(e)
                        )
                    )

            elif isinstance(schema[prop], dict):
                if prop not in cfg:
                    cfg[prop] = {}
                if cfg[prop] is None:
                    cfg[prop] = {}

                ConfigParserBase.validate(
                    schema[prop], cfg[prop],
                    prop if path == "" else "{}.{}".format(path, prop)
                )

            else:
                raise NotImplementedError("prop: {}, path: {}".format(prop, path))

        if errors:
            raise ConfigError()

    def parse(self):
        """
        Contents of cfg dict is updated/normalized by validators.
        """
        raise NotImplementedError()

def convert_next_hop_policy(cfg):
    if not isinstance(cfg.get("filtering", None), dict):
        return
    if "next_hop_policy" in cfg["filtering"]:
        if "next_hop" in cfg["filtering"]:
            raise ConfigError(
                "Can't use the new 'next_hop' and the old 'next_hop_policy' "
                "statements simultaneously for NEXT_HOP policy configuration"
            )

        cfg["filtering"]["next_hop"] = {
            "policy": cfg["filtering"]["next_hop_policy"]
        }
        del cfg["filtering"]["next_hop_policy"]

def convert_maxprefix_peeringdb(cfg):
    if not isinstance(cfg.get("filtering", None), dict):
        return
    if "max_prefix" not in cfg["filtering"]:
        return
    if not isinstance(cfg["filtering"]["max_prefix"], dict):
        return
    if "peering_db" in cfg["filtering"]["max_prefix"]:
        peering_db = cfg["filtering"]["max_prefix"]["peering_db"]
        if isinstance(peering_db, dict):
            return
        cfg["filtering"]["max_prefix"]["peering_db"] = {
            "enabled": peering_db
        }

def convert_filtering_rpki(cfg):
    if not isinstance(cfg.get("filtering", None), dict):
        return
    if "rpki" not in cfg["filtering"]:
        return
    if "rpki_bgp_origin_validation" in cfg["filtering"]:
        raise ConfigError(
            "A conflict due to a deprecated syntax exists: "
            "filtering.rpki and filtering.rpki_bgp_origin_validation "
            "are both configured."
        )
    cfg["filtering"]["rpki_bgp_origin_validation"] = \
        cfg["filtering"].pop("rpki")

def build_rpki_roas(cfg):
    """Build rpki_roas.

    Also used to identify those cases where filtering.rpki is enable
    (and then 'rtrlib' is implicitly used for ROAs collection) and
    filtering.irrdb.rpki_roas_as_route_objects is enabled and source
    is set to 'ripe-rpki-validator-cache'.
    """

    if not isinstance(cfg.get("filtering", None), dict):
        return

    rpki_roas = {}

    def from_rpki_roas_as_route_objects():
        """Return (used [bool], source)"""
        irrdb = cfg["filtering"].get("irrdb", None)
        if not isinstance(irrdb, dict):
            return False, None

        roas_as_route_objects = irrdb.get("use_rpki_roas_as_route_objects",
                                          None)
        if not isinstance(roas_as_route_objects, dict):
            return False, None

        if roas_as_route_objects.get("enabled", False) is not True:
            return False, None

        for k in ("source", "ripe_rpki_validator_url",
                  "allowed_trust_anchors"):
            if k in roas_as_route_objects:
                rpki_roas[k] = roas_as_route_objects.pop(k)

        return True, rpki_roas.get("source", None)

    def from_rpki():
        """Return (used [bool], source)"""
        rpki = cfg["filtering"].get("rpki", None)
        if not isinstance(rpki, dict):
            return False, None

        if rpki.get("enabled", False) is not True:
            return False, None

        rpki_roas["source"] = "rtrlib"

        return True, "rtrlib"

    roas_as_routes_used, roas_as_routes_src = from_rpki_roas_as_route_objects()
    rpki_used, rpki_src = from_rpki()

    if roas_as_routes_used and rpki_used and \
        (roas_as_routes_src or rpki_src) != rpki_src:
        raise ConfigError(
            "A deprecated syntax triggered an issue with the configuration "
            "of RPKI BGP Origin Validation (filtering.rpki) and ROAs-as-route-"
            "objects (filtering.irrdb.rpki_roas_as_route_objects). "
            "The former uses rtrlib as source for ROAs, while the "
            "latter is configured to use the RIPE RPKI Validator "
            "cache file. "
            "To fix this issue, convert them to the new syntax and configure "
            "the desired ROAs source within the 'rpki_roas' section. "
            "Please refer to the general.yml file distributed with the tool "
            "for the proper syntax to use."
        )

    if rpki_roas:
        if "rpki_roas" in cfg:
            raise ConfigError(
                "A conflict due to a deprecated syntax exists: "
                "please check rpki_roas, filtering.rpki and "
                "filtering.irrdb.rpki_roas_as_route_objects."
            )
        cfg["rpki_roas"] = rpki_roas

def convert_deprecated(cfg):
    if not cfg:
        return
    if not isinstance(cfg, dict):
        return

    # Convert next_hop_policy (< v0.6.0) into the new format
    convert_next_hop_policy(cfg)

    # Convert max_prefix.peering_db (< v0.13.0) into the new format
    convert_maxprefix_peeringdb(cfg)

    # Build cfg.rpki_roas (< v0.17.0) from rpki and roas_as_route_objects
    build_rpki_roas(cfg)

    # Convert filtering.rpki (< v0.17.0) into the new format
    convert_filtering_rpki(cfg)
